/****************************************************************/
/*              Reductions Application Suite                    */
/****************************************************************/

#include<stdio.h>
#include<stdlib.h>
#include<assert.h>
#include "timers.h"

volatile int lock;
#define LOCK_BUSY 1
#define LOCK_FREE 0

inline static void _lock_acquire()
{
	if ( (lock == LOCK_FREE) && !__sync_lock_test_and_set( &lock , LOCK_BUSY ) ) return;
spin:
    while ( lock == LOCK_BUSY ) {}
	if ( __sync_lock_test_and_set( &lock, LOCK_BUSY ) ) goto spin;
	return;
}

inline static void _lock_release()
{
	__sync_lock_release( &lock );
}

int max_height, max_childs, final_depth;
float creation_prob;

struct node_t
{
    int size;
    struct node_t **children;
};
int max(int x, int y)
{
    return x > y ? x : y;
}

int height_serial(struct node_t* node)
{
    if (node == 0)
        return 0;

    int i = 0;
    int max_height_childs = 0;
    while (i < node->size)
    {
        max_height_childs = max(max_height_childs, height_serial(node->children[i]));
        i++;
    }

    return max_height_childs + 1;
}

int height_ompss(struct node_t* node, int level)
{
    if (node == 0)
        return 0;

    int i = 0;
    int max_height_childs = 0;
    while (i < node->size)
    {
        #pragma omp task final(final_depth < level)
        {
            int tmp = height_ompss(node->children[i], level+1);
            _lock_acquire();
            if (max_height_childs < tmp)
            {
                max_height_childs = tmp;
            }
            _lock_release();
        }
        i++;
    }
#pragma omp taskwait

    return max_height_childs + 1;
}



struct node_t* new_tree_rec(int count)
{
    double double_rand = drand48();

    // do not create if the height of the current tree is equal to the max_height
    if (count == max_height
            // probability of create a new node 
            || double_rand > creation_prob)
    {
        return 0;
    }

    int num_childs = (rand() % max_childs) + 1;
    struct node_t* new_node = malloc(sizeof(struct node_t));
    struct node_t** children = malloc(sizeof(struct node_t*) * num_childs);
    new_node->size = num_childs;
    new_node->children = children;
    int i = 0;
    while (i < num_childs)
    {
        children[i] = new_tree_rec(count + 1);
        i++;
    }

    return new_node;
}

struct node_t* new_tree()
{
    return new_tree_rec(0);
}

void free_tree(struct node_t* node)
{
    if (node != 0)
    {
        int i = 0;
        while (i < node->size)
        {
            free_tree(node->children[i]);
            i++;
        }

        free(node->children);
        free(node);
    }
}


int main(int argc, char *argv[])
{
    if(argc != 6)
    {
        printf("./max_height_tree seed creation_prob max_height max_childs_per_node final_depth\n");
        printf("where:\n"
                "-seed: value used to initialize the GNA,\n"
                "-creation_prob: probability of create a new node (between 0..1),\n"
                "-max_height: value used to determine the max tree height,\n"
                "-max_childs_per_node: value used to determine the max number of childs per node,\n"
                "-final_depth: how many levels should be solved in parallel\n");
        exit(1);
    }
    int seed = atoi(argv[1]);
    creation_prob = atof(argv[2]);
    max_height = atoi(argv[3]);
    max_childs = atoi(argv[4]);
    final_depth= atoi(argv[5]);

    srand(seed);
    srand48(seed);

    lock = 0;

    struct node_t * root = new_tree();

    int real_height = height_serial(root);
    timer_start(0);
    int computed_height = height_ompss(root, /* current level */ 0);
    timer_stop(0);
    free_tree(root);

    printf("%.6f, %i, %i, %.6f, %i, %i, %i, %i\n", timer_read(0), \
       		computed_height,
       		seed,\
       		creation_prob,\
       		max_height,\
       		max_childs,\
       		omp_get_num_threads(), \
       		final_depth);

    //printf("real_height: %d == computed_height: %d ?\n", real_height, computed_height);
    assert(real_height == computed_height);
}
